// pdp1dump.c: dump PDP-1 RIM and BIN format paper tape images
// http://www.brouhaha.com/~eric/software/pdp1dump/
// $Id: pdp1dump.c 2 2004-07-07 22:15:25Z eric $
// Copyright 2004 Eric Smith <eric@brouhaha.com>

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License version
// 2 as published by the Free Software Foundation.  Note that
// permission is not granted to redistribute this program under the
// terms of any other version of the General Public License.

// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

// Usage:
//    pdp1dump [options] [file]

// If file is not specified, standard in is used.

// If "-q" option is given, data is not output, but BIN record addresses
// and checksum verification messages are still output.

// If "-m" option is given, deal with tapes that consist of a series
// of concatenated RIM images, such as MAINDEC 1 (Instruction Test).
// Otherwise, the tape is asssumed to consist of one RIM record and a
// series of BIN records.


#include <stdio.h>
#include <string.h>


#define RIM 0
#define BIN 1


// options:
int mult_rim;
int quiet;


int get_tape_char (FILE *f)
{
  int c;

  do
    {
      c = fgetc (f);
      if (c == EOF)
	return (EOF);
    }
  while (! (c & 0x80));
  return (c & 077);
}


int get_tape_word (FILE *f)
{
  int c;
  int i;
  int w = 0;

  for (i = 0; i < 3; i++)
    {
      c = get_tape_char (f);
      if (c == EOF)
	return (EOF);
      w = (w << 6) + c;
    }

  return (w);
}


int add_ones_comp (int a, int b)
{
  int r;
  r = a + b;
  if (r & (1 << 18))
    {
      r &= ((1 << 18) - 1);
      r++;
    }
  return (r);
}


int dump_rim_record (FILE *f)
{
  int addr, data;
  int af = 1;
  int p = 0;
  int hdr = 0;

  int prev_addr;

  for (;;)
    {
      addr = get_tape_word (f);
      if (addr == EOF)
	return (0);

      if ((addr >> 12) == 060)
	{
	  printf ("\nRIM start addr %04o\n", addr & 07777);
	  return (1);
	}

      if ((addr >> 12) != 032)
	{
	  printf ("\nbad RIM format addr word %06o\n", addr);
	  return (0);
	}

      addr &= 07777;

      if (! hdr)
	{
	  printf ("RIM format record starts at address %04o\n", addr);
	  hdr = 1;
	}

      data = get_tape_word (f);
      if (data == EOF)
	return (0);

      if (! quiet)
	{
	  if (af || (addr != (prev_addr + 1)) || (p > 7))
	    {
	      if (p)
		{
		  printf ("\n");
		  p = 0;
		}
	      printf ("  %04o:", addr);
	      af = 0;
	    }

	  printf ("  %06o", data);
	  p++;
	}

      prev_addr = addr;

    }
}


int dump_bin_record (FILE *f)
{
  int addr, end;
  int data;
  int checksum;
  int af;

  addr = get_tape_word (f);
  if (addr == EOF)
    return (0);

  if ((addr >> 12) == 060)
    {
      printf ("BIN start record, addr %04o\n", addr & 07777);
      return (1);
    }

  if ((addr >> 12) != 032)
    {
      printf ("bad BIN format address %06o\n", addr);
      return (0);
    }

  checksum = addr;

  addr &= 07777;

  end = get_tape_word (f);
  if (end == EOF)
    return (0);

  if ((end >> 12) != 032)
    {
      printf ("bad BIN format end address word %06o\n", end);
      return (0);
    }

  end &= 07777;

  printf ("BIN format record, first addr %04o, last addr %04o\n", addr, end - 1);

  af = 1;

  while (addr < end)
    {
      data = get_tape_word (f);
      if (data == EOF)
	return (0);

      if (! quiet)
	{
	  if (af || ((addr & 7) == 0))
	    {
	      printf ("  %4o:", addr);
	      af = 0;
	    }

	  printf ("  %06o", data);
	  if ((addr & 7) == 7)
	    printf ("\n");
	}

      checksum = add_ones_comp (checksum, data);
      addr++;
    }

  checksum = add_ones_comp (checksum, (032 << 12) + addr);

  if (! quiet)
    {
      if ((addr & 7) != 0)
	printf ("\n");
    }

  data = get_tape_word (f);
  if (data == EOF)
    return (0);

  if (data != checksum)
    printf ("checksum error: tape %06o, computed %06o\n", data, checksum);
  else
    printf ("checksum good\n");

  return (1);
}


int main(int argc, char *argv[])
{
  FILE *f = NULL;
  char *fn = NULL;

  int ok;
  int mode;

  mult_rim = 0;
  quiet = 0;

  while (--argc)
    {
      argv++;
      if ((argv [0][0] == '-') && (argv [0][1] != '\0'))
	{
	  if (strcmp (argv [0], "-m") == 0)
	    mult_rim = 1;
	  else if (strcmp (argv [0], "-q") == 0)
	    quiet = 1;
	  else
	    {
	      fprintf (stderr, "Unrecognized option '%s'\n", argv [0]);
	      return (1);
	    }
	}
      else
	{
	  if (! fn)
	    fn = argv [0];
	  else
	    {
	      fprintf (stderr, "only one file may be specified\n");
	      return (1);
	    }
	}
    }

  if ((fn) && (strcmp (fn, "-") != 0))
    {
      f = fopen (fn, "rb");
      if (! f)
	{
	  fprintf (stderr, "Error opening input file '%s'\n", fn);
	  return (2);
	}
    }
  else
    f = stdin;

  mode = RIM;

  ok = 1;

  while (ok)
    {
      switch (mode)
	{
	case RIM:
	  ok = dump_rim_record (f);
	  if (! mult_rim)
	    mode = BIN;
	  break;
	case BIN:
	  ok = dump_bin_record (f);
	  break;
	}
      printf ("\n");
    }

  return (0);
}

